Attribute VB_Name = "modDatabase"
Option Explicit

Const dbStructureFile As String = "dbADO.txt"

Global cat As ADOX.Catalog, cnn As ADODB.Connection

Public Enum ResultCodes
  rcUnknown = 0
  rcOK = 1
  rcCancel = 2
  rcError = 3
End Enum

Public Function InitTables(ByVal TableName As String) As ResultCodes
Dim FN As Long, F As String, L As String, I As Long, fInsideMode As Boolean
Const BeginWord As String = "BEGIN", EndWord As String = "END"
Const TableWord As String = "TABLE"
For I = cat.Tables.Count - 1 To 0 Step -1
  If cat.Tables(I).Name = TableName Then Exit For
Next I
If I >= 0 Then
  InitTables = rcCancel
  Exit Function
End If
F = App.Path: If Right$(F, 1) = "\" Then F = F & dbStructureFile Else F = F & "\" & dbStructureFile
If Len(Dir$(F)) = 0 Then
  InitTables = rcError
  Exit Function
End If
FN = FreeFile
fInsideMode = False
L = vbNullString
Open F For Input Access Read Lock Write As #FN
  Do Until EOF(FN)
    Line Input #FN, F
    If fInsideMode Then
      L = L & F & vbNewLine
      F = UCase$(Trim$(F))
      If Left$(F, Len(EndWord) + 1) = EndWord & vbSpace Then
        If Left$(F, Len(EndWord) + Len(TableWord) + 1) = EndWord & vbSpace & TableWord Then
          L = Left$(L, Len(L) - Len(F) - Len(vbNewLine))
          Exit Do
        End If
      End If
    Else
      F = UCase$(Trim$(F))
      If Left$(F, Len(BeginWord) + 1) = BeginWord & vbSpace Then
        If Left$(F, Len(BeginWord) + Len(TableWord) + 2) = BeginWord & vbSpace & TableWord & vbSpace Then
          F = Mid$(F, Len(BeginWord) + Len(TableWord) + 3)
          If InStr(F, vbSpace) > 0 Then F = Left$(F, InStr(F, vbSpace) - 1)
          If F = UCase$(TableName) Then fInsideMode = True
        End If
      End If
    End If
  Loop
Close #FN
If Not fInsideMode Then
  InitTables = rcError
  Exit Function
End If
Call CreateTable(TableName, L)
InitTables = rcOK
End Function

Public Function CreateTable(ByVal TableName As String, ByVal StructureString As String) As ResultCodes
Dim fInsideMode As Boolean, N As Long, I0 As Long, I As Long, S As String, S0 As String
Dim faCounter As Boolean, faFixed As Boolean, faNullable As Boolean, faNumeric As Boolean, faText As Boolean
Dim fld() As String, V() As String, Hidden As Boolean, FV As Variant
Dim objTable As ADOX.Table, objColumn As ADOX.Column
Const CommentChar As String = "#", SeparatorChar As String = "|"
Const BeginWord As String = "BEGIN VALUES", EndWord As String = "END VALUES"
Const FieldWord As String = "FIELD", IndexWord As String = "INDEX"
Const HiddenWord As String = "HIDDEN"
ReDim fld(0), idx(0), V(0)
fInsideMode = False
I0 = 1
Do
  I = InStr(I0, StructureString, vbNewLine)
  If I = 0 Then Exit Do
  S = Trim$(Mid$(StructureString, I0, I - I0))
  If fInsideMode Then
    If UCase$(Left$(S, Len(EndWord))) = EndWord Then
      fInsideMode = False
    Else
      ReDim Preserve V(0 To UBound(V) + 1)
      V(UBound(V)) = S
    End If
  Else
    If InStr(S, CommentChar) > 0 Then S = RTrim$(Left$(S, InStr(S, CommentChar) - 1))
    I0 = 0
    While I0 <> Len(S)
      I0 = Len(S)
      S = Replace(S, vbSpace2, vbSpace)
    Wend
    S = UCase$(S)
    If Left$(S, Len(BeginWord)) = BeginWord Then
      fInsideMode = True
    Else
      If Left$(S, Len(HiddenWord)) = HiddenWord Then
        Hidden = True
      ElseIf Left$(S, Len(FieldWord) + 1) = FieldWord & vbSpace Then
        ReDim Preserve fld(0 To UBound(fld) + 1)
        fld(UBound(fld)) = Mid$(S, Len(FieldWord) + 2)
      ElseIf Left$(S, Len(IndexWord) + 1) = IndexWord & vbSpace Then
        ReDim Preserve idx(0 To UBound(idx) + 1)
        idx(UBound(idx)) = Mid$(S, Len(IndexWord) + 2)
      End If
    End If
  End If
  I0 = I + Len(vbNewLine)
Loop
S = vbNullString
cnn.BeginTrans
Set objTable = New ADOX.Table
objTable.Name = TableName
'objTable.Attributes = objTableDef.Attributes Or IIf(Hidden, dbHiddenObject, 0&)
For N = 1 To UBound(fld)
  faCounter = False
  faFixed = False
  faNullable = False
  faNumeric = False
  faText = False
  S = vbNullString
  fld(0) = vbNullString
  I0 = 0
  I = InStr(fld(N), SeparatorChar)
  If I > 0 Then
    fld(0) = Trim$(Left$(fld(N), I - 1))
    fld(N) = Trim$(Mid$(fld(N), I + Len(SeparatorChar)))
  End If
  I = InStr(fld(N), SeparatorChar)
  If I > 0 Then
    S = Trim$(Mid$(fld(N), I + Len(SeparatorChar)))
    fld(N) = Trim$(Left$(fld(N), I - 1))
  End If
  If Len(S) > 0 Then
    I = InStr(S, "-")
    If I > 0 Then
      faFixed = True
      Mid$(S, I, 1) = "*"
    End If
    I = InStr(S, "*")
    If I > 0 Then
      S0 = Mid$(S, I + 1)
      S = Left$(S, I - 1)
      I = InStr(S0, ".")
      If I = 0 Then
        I = Val(S0)
      Else
        I0 = Val(Mid$(S0, I + 1))
        I = Val(Left$(S0, I - 1))
      End If
    End If
  End If
  faNullable = (InStr(fld(0), "R") = 0)
  If fld(0) = "+" Then
    faCounter = True
    faNullable = False
    S = "LONG"
  End If
  Set objColumn = New ADOX.Column
  objColumn.Name = fld(N)
  Select Case S
    Case "BOOL", "BOOLEAN"
      objColumn.Type = adBoolean
    Case "BYTE"
      objColumn.Type = adUnsignedTinyInt
      faNumeric = True
    Case "INT", "INTEGER", "LONG"
      objColumn.Type = adInteger
      faNumeric = True
    Case "SNG", "SINGLE"
      objColumn.Type = adSingle
      faNumeric = True
    Case "DBL", "DOUBLE"
      objColumn.Type = adDouble
      faNumeric = True
    Case "DEC", "DECIMAL"
      objColumn.Type = adDecimal
      faNumeric = True
    Case "NUM", "NUMERIC"
      objColumn.Type = adNumeric
      faNumeric = True
    Case "CURRENCY"
      objColumn.Type = adCurrency
      faNumeric = True
    Case "DATETIME"
      objColumn.Type = adDate
    Case "DATE"
      objColumn.Type = adDBDate
    Case "TIME"
      objColumn.Type = adDBTime
    Case "TIMESTAMP"
      objColumn.Type = adDBTimeStamp
    Case "BSTR"
      objColumn.Type = adBSTR
      faText = True
    Case "CHAR", "TEXT", "STR", "STRING"
      If faFixed Then
        objColumn.Type = adChar
      Else
        objColumn.Type = adVarChar
      End If
      faText = True
      'Forcing to unicode
      If faFixed Then
        objColumn.Type = adWChar
      Else
        objColumn.Type = adVarWChar
      End If
    Case "MEMO"
      objColumn.Type = adLongVarChar
      faFixed = False
      faText = True
      'Forcing to unicode
      objColumn.Type = adLongVarWChar
    Case "WCHAR", "WTEXT", "WSTR", "WSTRING", "WIDECHAR", "WIDETEXT", "WIDESTR", "WIDESTRING", "WIDE CHAR", "WIDE TEXT", "WIDE STRING"
      If faFixed Then
        objColumn.Type = adWChar
      Else
        objColumn.Type = adVarWChar
      End If
      faText = True
    Case "WMEMO", "WIDEMEMO", "WIDE MEMO", "UNIMEMO", "UNI MEMO", "UNICODE"
      objColumn.Type = adLongVarWChar
      faFixed = False
      faText = True
  End Select
  If faNumeric Then
    objColumn.Precision = I
    objColumn.NumericScale = I0
  End If
  If faText Then
    objColumn.DefinedSize = I
  End If
  objColumn.Attributes = IIf(faFixed, adColFixed, 0) Or IIf(faNullable, adColNullable, 0)
  'objColumn.Properties("AutoIncrement") = faCounter
  objTable.Columns.Append objColumn
  Set objColumn = Nothing
Next N
cat.Tables.Append objTable
Set objTable = Nothing
cnn.CommitTrans
Call RebuildIndexes(TableName, StructureString)
'LoadRowData V()
Erase fld(), V()
CreateTable = rcOK
End Function

Public Function RebuildIndexes(ByVal TableName As String, ByVal StructureString As String) As ResultCodes
Dim fInsideMode As Boolean, N As Long, I0 As Long, I As Long, S As String, S0 As String
Dim iaPrimary As Boolean, iaUnique As Boolean, iaRequired As Boolean, iaNulls As Boolean
Dim idx() As String, fld() As String, FV As Variant, objIndex As ADOX.Index
Const CommentChar As String = "#", SeparatorChar As String = "|"
Const BeginWord As String = "BEGIN VALUES", EndWord As String = "END VALUES"
Const FieldWord As String = "FIELD", IndexWord As String = "INDEX"
ReDim idx(0), fld(0)
fInsideMode = False
I0 = 1
Do
  I = InStr(I0, StructureString, vbNewLine)
  If I = 0 Then Exit Do
  S = Trim$(Mid$(StructureString, I0, I - I0))
  If fInsideMode Then
    If UCase$(Left$(S, Len(EndWord))) = EndWord Then
      fInsideMode = False
    End If
  Else
    If InStr(S, CommentChar) > 0 Then S = RTrim$(Left$(S, InStr(S, CommentChar) - 1))
    I0 = 0
    While I0 <> Len(S)
      I0 = Len(S)
      S = Replace(S, vbSpace2, vbSpace)
    Wend
    S = UCase$(S)
    If Left$(S, Len(BeginWord)) = BeginWord Then
      fInsideMode = True
    Else
      If Left$(S, Len(IndexWord) + 1) = IndexWord & vbSpace Then
        ReDim Preserve idx(0 To UBound(idx) + 1)
        idx(UBound(idx)) = Mid$(S, Len(IndexWord) + 2)
      End If
    End If
  End If
  I0 = I + Len(vbNewLine)
Loop
S = vbNullString
For I = cat.Tables(TableName).Indexes.Count - 1 To 0 Step -1
  cat.Tables(TableName).Indexes.Delete I
Next I
For N = 1 To UBound(idx)
  S = vbNullString
  idx(0) = vbNullString
  I0 = 0
  I = InStr(idx(N), SeparatorChar)
  If I > 0 Then
    idx(0) = Trim$(Left$(idx(N), I - 1))
    idx(N) = Trim$(Mid$(idx(N), I + Len(SeparatorChar)))
  End If
  I = InStr(idx(N), SeparatorChar)
  If I > 0 Then
    S = Trim$(Mid$(idx(N), I + Len(SeparatorChar)))
    idx(N) = Trim$(Left$(idx(N), I - 1))
  End If
  iaPrimary = (InStr(idx(0), "*") > 0)
  iaUnique = (InStr(idx(0), "U") > 0)
  iaRequired = (InStr(idx(0), "R") > 0)
  iaNulls = (InStr(idx(0), "N") > 0)
  If iaPrimary Then
    iaUnique = True
    iaRequired = True
    iaNulls = False
  End If
  If iaUnique Then
    iaRequired = True
  End If
  fld() = Split(S, ";")
  Set objIndex = New ADOX.Index
  objIndex.Name = idx(N)
  objIndex.PrimaryKey = iaPrimary
  objIndex.Unique = iaUnique
  objIndex.IndexNulls = IIf(iaRequired, _
                            adIndexNullsDisallow, _
                            IIf(iaNulls, adIndexNullsAllow, adIndexNullsIgnore))
  For I = LBound(fld) To UBound(fld)
    If Left$(fld(I), 1) = "+" Then
      S = " Asc"
      fld(I) = Mid$(fld(I), 2)
    ElseIf Left$(fld(I), 1) = "-" Then
      S = " Desc"
      fld(I) = Mid$(fld(I), 2)
    Else
      S = vbNullString
    End If
    If InStr(fld(I), vbSpace) > 0 Then
      If Left$(fld(I), 1) <> "[" And Right$(fld(I), 2) <> "]" Then fld(I) = "[" & fld(I) & "]"
    End If
    objIndex.Columns.Append fld(I)
  Next I
  cat.Tables(TableName).Indexes.Append objIndex
  Set objIndex = Nothing
Next N
Erase idx(), fld()
RebuildIndexes = rcOK
End Function

Public Function LoadRowData(ByVal TableName As String, Values() As String) As ResultCodes
Dim rs As ADODB.Recordset, C As Long, I As Long, S As String, F() As String, V() As String
Const QC As String = """"
If UBound(Values) = 0 Then
  LoadRowData = rcOK
  Exit Function
End If
Set rs = New ADODB.Recordset
rs.Open TableName, cnn, adOpenForwardOnly, adLockOptimistic, adCmdTable
ReDim F(0 To rs.Fields.Count - 1)
For I = 0 To rs.Fields.Count - 1
  F(I) = rs.Fields(I).Name
Next I
cnn.BeginTrans
For I = 1 To UBound(Values)
  S = Values(I)
  V() = Split(S, vbNullChar)
  rs.Update F(), V()
Next I
rs.Close
Set rs = Nothing
'If UBound(V) > 0 Then
'  Set rs = db.OpenRecordset(TableName, dbOpenTable)
'  For I = 1 To UBound(V)
'    rs.AddNew
'    For I0 = 0 To rs.Fields.Count - 1
'      FV = Null
'      rs.Fields(I).Value = FV
'    Next I0
'    rs.Update
'  Next I
'  rs.Close: Set rs = Nothing
'End If
cnn.CommitTrans
LoadRowData = rcOK
End Function
