Attribute VB_Name = "modDatabase"
Option Explicit

Global db As DAO.Database

Public Enum ResultCodes
  rcUnknown = 0
  rcOK = 1
  rcCancel = 2
  rcError = 3
End Enum

Public Function InitTables(ByVal TableName As String) As ResultCodes
Dim TD As TableDef, I As Long
For I = 0 To db.TableDefs.Count - 1
  If db.TableDefs(I).Name = TableName Then Exit For
Next I
If I < db.TableDefs.Count Then Exit Function
Select Case TableName
  Case "OPTIONS"
    Set TD = CreateTableDef(TableName, "*|KEY|GROUP;NAME;INDEX", _
    "RZ|GROUP|Text*250", "RZ|NAME|Text*250", "R|INDEX|Long", "R|TYPE|Long", "Z|VALUE|Memo")
  Case "USERS"
    Set TD = CreateTableDef(TableName, Array("*|LOGIN|LOGIN", "NAME|NAME"), _
    "R|LOGIN|Text*50", "RZ|NAME|Text*250", _
    "RZ|NAME0|Text*100", "RZ|NAME1|Text*100", "RZ|NAME2|Text*100", _
    "RZ|TITLE|Text*100", "Z|PASSWORD|Text*40")
  Case "REF_CITY"
    Set TD = CreateTableDef(TableName, Array("*|KEY|ID", "U|NAME|TYPE;NAME", "N|POSTINDEX|POSTINDEX"), _
    "ID|+", "R|NAME|Text*250", "RZ|TYPE|Text*20", "|POSTINDEX|Text*10")
  Case "REF_STREET"
    Set TD = CreateTableDef(TableName, Array("*|KEY|ID", "U|STREET|CITY;TYPE;NAME", "NAME|TYPE|NAME", "CITY|CITY"), _
    "ID|+", "R|NAME|Text*250", "RZ|TYPE|Text*20", "R|CITY|Long")
  Case Else
    InitTables = rcError
    Exit Function
End Select
db.TableDefs.Append TD
Set TD = Nothing
InitTables = rcOK
End Function

Public Function CreateTableDef(Name As String, Indexes As Variant, ParamArray Fields() As Variant) As TableDef
'Indexes - string or string array
'  Item: [[*][U][N]|]Name|[+|-]field ...
'    * - Primary
'    U - Unique (default for primary)
'    N - Ignore nulls
'    Field  - field(s) numbers
'Fields: [[R][Z]|]Name|Type[*Size]
'  R - Required field (default)
'  Z - Allow zero length
'  Name - Field name
'  Type - Field type
'    +     - Long Auto Increment (counter)
'    BigInt- BigInt
'    Bin   - Binary
'    Bool  - Boolean
'    Byte  - Byte
'    Char  - Char
'    Cur   - Currency
'    Date  - Date
'    Dec   - Decimal
'    Dbl   - Double
'    Float - Float
'    GUID  - GUID
'    Int   - Integer
'    Long  - Long
'    LBin  - LongBinary
'    Memo  - Memo
'    Num   - Numeric
'    Sng   - Single
'    Text  - Text
'    Time  - Time
'  Size - Field size (in bytes for Text and Char)
Dim TD As TableDef, fld As Field, idx As Index
Dim C As Long, N As Long, I As Long, Z As String
Dim S0 As String, S1 As String, S2 As String
If VarType(Indexes) = vbString Then
  Z = Indexes
  ReDim Indexes(0)
  Indexes(0) = Z
End If
Set TD = New TableDef
TD.Name = Name
For C = LBound(Fields) To UBound(Fields)
  If VarType(Fields(C)) = vbString Then Z = Fields(C) Else Z = ""
  If Len(Z) > 0 Then
    If InStr(Z, "|") = 0 Then
      S0 = Z
      S1 = "+"
      S2 = "R"
    Else
      I = InStrRev(Z, "|")
      S1 = Mid$(Z, I + 1)
      Z = Left$(Z, I - 1)
      If InStr(Z, "|") = 0 Then
        S0 = Z
        S2 = "R"
      Else
        I = InStrRev(Z, "|")
        S0 = Mid$(Z, I + 1)
        S2 = Left$(Z, I - 1)
      End If
    End If
    S1 = UCase$(S1): S2 = UCase$(S2)
    Set fld = New Field
    fld.Name = S0
    If InStr(S1, "-") > 0 Then
      Mid$(S1, InStr(S1, "-"), 1) = "*"
      fld.Attributes = fld.Attributes Or dbFixedField
    End If
    If InStr(S1, "*") > 0 Then
      I = Val(Mid$(S1, InStr(S1, "*") + 1))
      Z = UCase$(Left$(S1, InStr(S1, "*") - 1))
    Else
      I = 0
      Z = S1
    End If
    Select Case Z
      Case "+"
        fld.Type = dbLong
        fld.Attributes = fld.Attributes Or dbAutoIncrField
      Case "BIGINT", "BIG INT"
        fld.Type = dbBigInt
      Case "BIN", "BINARY"
        fld.Type = dbBinary
      Case "BOOL", "BOOLEAN"
        fld.Type = dbBoolean
      Case "BYTE"
        fld.Type = dbByte
      Case "CHAR", "CHAR"
        fld.Type = dbChar
      Case "CUR", "CURRENCY"
        fld.Type = dbCurrency
      Case "DATE", "DATE"
        fld.Type = dbDate
      Case "DEC", "DECIMAL"
        fld.Type = dbDecimal
      Case "DBL", "DOUBLE"
        fld.Type = dbDouble
      Case "FLOAT"
        fld.Type = dbFloat
      Case "GUID"
        fld.Type = dbGUID
      Case "INT", "INTEGER"
        fld.Type = dbInteger
      Case "LONG", "LONG"
        fld.Type = dbLong
      Case "LBIN", "LONGBINARY", "LONG BINARY"
        fld.Type = dbLongBinary
      Case "MEMO"
        fld.Type = dbMemo
      Case "NUM", "NUMERIC"
        fld.Type = dbNumeric
      Case "SNG", "SINGLE"
        fld.Type = dbSingle
      Case "TEXT", "TXT", "STRING", "STR"
        fld.Type = dbText
      Case "TIME"
        fld.Type = dbTime
    End Select
    If I > 0 Then fld.Size = I
    fld.Required = (InStr(S2, "R") > 0)
    Select Case fld.Type
      Case dbText, dbMemo
        fld.AllowZeroLength = (InStr(S2, "Z") > 0)
    End Select
    TD.Fields.Append fld
    Set fld = Nothing
  End If
Next C
For C = LBound(Indexes) To UBound(Indexes)
  If VarType(Indexes(C)) = vbString Then Z = Indexes(C) Else Z = ""
  If Len(Z) > 0 Then
    If InStr(Z, "|") = 0 Then
      S0 = Z
      S1 = ""
      S2 = ""
    Else
      I = InStrRev(Z, "|")
      S1 = Mid$(Z, I + 1)
      Z = Left$(Z, I - 1)
      If InStr(Z, "|") = 0 Then
        S0 = Z
      Else
        I = InStrRev(Z, "|")
        S0 = Mid$(Z, I + 1)
        S2 = Left$(Z, I - 1)
      End If
    End If
    If InStr(S2, "U") = 0 And InStr(S2, "*") > 0 Then S2 = S2 + "U"
    Set idx = New Index
    idx.Name = S0
    idx.Primary = (InStr(S2, "*") > 0)
    idx.Unique = (InStr(S2, "U") > 0)
    idx.IgnoreNulls = (InStr(S2, "N") > 0)
    N = UBound(Split(S1, ";")) - LBound(Split(S1, ";")) + 1
    S0 = ""
    For I = 1 To N
      Z = Split(S1, ";")(I - 1)
      If Left$(Z, 1) = "+" Then
        S2 = " Asc"
        Z = Mid$(Z, 2)
      ElseIf Left$(Z, 1) = "-" Then
        S2 = " Desc"
        Z = Mid$(Z, 2)
      Else
        S2 = ""
      End If
      If InStr(Z, " ") > 0 Then
        If Left$(Z, 1) <> "[" And Right$(Z, 2) <> "]" Then Z = "[" & Z & "]"
      End If
      If I > 1 Then S0 = S0 & ";"
      S0 = S0 & Z & S2
    Next I
    idx.Fields = S0
    TD.Indexes.Append idx
    Set idx = Nothing
  End If
Next C
Set CreateTableDef = TD
Set TD = Nothing
End Function
